/*-------------------------------------------------------------------------+
| Copyright 2013 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.quality.service;

import static java.util.Collections.emptyList;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getFirstChildWithType;

import java.util.ArrayList;
import java.util.List;
import java.util.Map.Entry;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.kernel.introspection.IIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectionItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectiveKernelService;
import org.fortiss.tooling.kernel.service.IKernelIntrospectionSystemService;
import org.fortiss.tooling.kernel.service.IMigrationService;
import org.fortiss.tooling.kernel.service.base.EObjectAwareServiceBase;

/**
 * This class implements the {@link IMigrationService} interface.
 * 
 * @author blaschke
 */
public class MetricService extends EObjectAwareServiceBase<IMetricProvider<?>>
		implements IIntrospectiveKernelService, IMetricService {
	/** The singleton instance. */
	private static final MetricService INSTANCE = new MetricService();

	/** Returns singleton instance of the service. */
	public static MetricService getInstance() {
		return INSTANCE;
	}

	/** The connector extension point ID. */
	private static final String EXTENSION_POINT_NAME =
			"org.fortiss.tooling.ext.quality.metricService";

	/** The connector configuration element name. */
	private static final String CONFIGURATION_ELEMENT_NAME = "metricService";

	/** The connector attribute name. */
	private static final String ATTRIBUTE_NAME = "metricService";

	/** */
	@Override
	public void startService() {
		IKernelIntrospectionSystemService.getInstance().registerService(this);
	}

	/** Registers the migration provider with the service. */
	@Override
	public void registerMetricProvider(IMetricProvider<?> provider, Class<?> modelElementClass) {
		addHandler(modelElementClass, provider);
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionDescription() {
		return getIntrospectionLabel() +
				"\n\nThis service allows the tracking of metrical development in modeling." +
				"\n the service outputs the current metric analysis of the model and stores" +
				"\n the data within logs for later history analysis." +
				"\n\nThe service extension point is '" + EXTENSION_POINT_NAME + "'.";
	}

	/** Get all suitable {@link IMetricProvider} for the given input. */
	@SuppressWarnings("unused")
	private List<IMetricProvider<?>> getAllMetricProviders(EObject input) {
		List<IMetricProvider<?>> providers = new ArrayList<>();
		for(Entry<Class<?>, List<IMetricProvider<?>>> migEntry : handlerMap.entrySet()) {
			if(getFirstChildWithType(input, migEntry.getKey()) != null) {
				providers.addAll(migEntry.getValue());
			}
		}
		return providers;
	}

	/**
	 * public List<IMetricProvider<? extends EObject>> performAllMetricAnalysis(EObject
	 * modelElement) {
	 * List<IMetricProvider<? extends EObject>> result =
	 * new LinkedList<IMetricProvider<? extends EObject>>();
	 * List<IMetricProvider<?>> handlers = getRegisteredHandlers(modelElement.getClass());
	 * if(handlers == null) {
	 * return null;
	 * }
	 * for(IMetricProvider<?> checker : handlers) {
	 * if(checker.isApplicable(modelElement)) {
	 * List<? extends IMetricProvider<? extends EObject>> metricsList =
	 * checker.apply(modelElement);
	 * if(metrics != null) {
	 * metricsList.addAll(metrics);
	 * }
	 * }
	 * }
	 * return result;
	 * }
	 */

	/** */
	@Override
	protected String getExtensionPointName() {
		return EXTENSION_POINT_NAME;
	}

	/** */
	@Override
	protected String getConfigurationElementName() {
		return CONFIGURATION_ELEMENT_NAME;
	}

	/** */
	@Override
	protected String getHandlerClassAttribute() {
		return ATTRIBUTE_NAME;
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionLabel() {
		return "Metric Service";
	}

	/** {@inheritDoc} */
	@Override
	public List<IIntrospectionItem> getIntrospectionItems() {
		return emptyList();
	}

	/** {@inheritDoc} */
	@Override
	public boolean showInIntrospectionNavigation() {
		return true;
	}

	@SuppressWarnings("javadoc")
	@Override
	public IIntrospectionDetailsItem getDetailsItem() {
		return null;
	}
}
