/*-------------------------------------------------------------------------+
| Copyright 2011 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.internal;

import static java.util.Collections.emptyList;
import static java.util.Collections.sort;
import static java.util.Collections.unmodifiableList;
import static org.conqat.lib.commons.collections.CollectionUtils.asUnmodifiable;
import static org.fortiss.tooling.kernel.utils.ExtensionPointUtils.getBundle;
import static org.fortiss.tooling.kernel.utils.ExtensionPointUtils.getConfigurationElements;
import static org.fortiss.tooling.kernel.utils.ExtensionPointUtils.loadClass;
import static org.fortiss.tooling.kernel.utils.LoggingUtils.error;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import org.conqat.lib.commons.collections.UnmodifiableList;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.kernel.ToolingKernelActivator;
import org.fortiss.tooling.kernel.extension.IPrototypeProvider;
import org.fortiss.tooling.kernel.extension.data.Prototype;
import org.fortiss.tooling.kernel.extension.data.PrototypeCategory;
import org.fortiss.tooling.kernel.introspection.IIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectionItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectiveKernelService;
import org.fortiss.tooling.kernel.introspection.items.PrototypeKISSDetailsItem;
import org.fortiss.tooling.kernel.service.IElementCompositorService;
import org.fortiss.tooling.kernel.service.IKernelIntrospectionSystemService;
import org.fortiss.tooling.kernel.service.IPrototypeService;
import org.fortiss.tooling.kernel.service.ITutorialService;
import org.osgi.framework.Bundle;

/**
 * This class implements the {@link IPrototypeService} interface.
 * 
 * @author hoelzl
 */
public class PrototypeService implements IPrototypeService, IIntrospectiveKernelService {
	/** The singleton instance. */
	private static final PrototypeService INSTANCE = new PrototypeService();

	/** Returns singleton instance of the service. */
	public static PrototypeService getInstance() {
		return INSTANCE;
	}

	/** The prototype provider extension point ID. */
	private static final String EXTENSION_POINT_NAME =
			"org.fortiss.tooling.kernel.modelPrototypeProvider";

	/** The prototype provider configuration element name. */
	private static final String CONFIGURATION_ELEMENT_NAME = "modelPrototypeProvider";

	/** Stores the registered prototype providers. */
	private final List<IPrototypeProvider> prototypeProviderList =
			new ArrayList<IPrototypeProvider>();
	/** The tutorial service instance. */
	private ITutorialService tutorialService;

	/** Cache for (expensive) {@link #getComposablePrototypes(Class)}. */
	private Map<Class<? extends EObject>, List<Prototype>> composablePrototypesCache =
			new HashMap<>();

	/** Initializes the service. */
	public void initializeService() {
		setupPrototypes();
	}

	/** Starts the service. */
	public void startService() {
		IKernelIntrospectionSystemService.getInstance().registerService(this);

		tutorialService = ITutorialService.getInstance();
		// never call unget, since prototype service never dies
	}

	/** Registers the given provider with the service. */
	@Override
	public void registerPrototypeProvider(IPrototypeProvider provider) {
		prototypeProviderList.add(provider);
	}

	/** Updates all prototypes by calling refreshCurrentPrototypes() for all providers. */
	public void updatePrototypes() {
		// cache needs to be cleared,
		// otherwise the following prototype changes would be ignored
		composablePrototypesCache.clear();

		List<IPrototypeProvider> providerList = getPrototypeProviders();
		for(IPrototypeProvider provider : providerList) {
			provider.refreshCurrentPrototypes();
		}
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionDescription() {
		return getIntrospectionLabel() + "\n\nThis service manages the model element prototypes." +
				"\nTo understand what a prototype is consider what happens when you drag and drop an element from the palette:" +
				"\nwhat happens internally is that we copy a reference element from the palette into the project." +
				"\nThis reference element is what we call a \"prototype\"" +
				"\nUsually model elements are composed of a couple of EObjects and have their attributes set to some predefined values." +
				"\nEach part of the modeling language(s) can register prototype provider classes, which then provide the model elements." +
				"\n\nThe service extension point is '" + EXTENSION_POINT_NAME + "'.";
	}

	/** Initializes the prototype list from plugin extensions. */
	private void setupPrototypes() {
		for(IConfigurationElement ce : getConfigurationElements(EXTENSION_POINT_NAME,
				CONFIGURATION_ELEMENT_NAME)) {
			Bundle bundle = getBundle(ce);
			try {
				Class<?> handlerClass = loadClass(ce.getAttribute("provider"), bundle);
				IPrototypeProvider provider =
						(IPrototypeProvider)handlerClass.getConstructor().newInstance();
				prototypeProviderList.add(provider);
			} catch(Exception ex) {
				error(ToolingKernelActivator.getDefault(), ex.getMessage(), ex);
			}
		}
	}

	/** {@inheritDoc} */
	@Override
	public UnmodifiableList<IPrototypeProvider> getPrototypeProviders() {
		return asUnmodifiable(prototypeProviderList);
	}

	/** {@inheritDoc} */
	@Override
	public UnmodifiableList<Prototype> getAllPrototypes() {
		if(tutorialService.isTutorialActive()) {
			return asUnmodifiable(getPrototypes(false).stream()
					.filter(proto -> tutorialService.prototypeActive(proto))
					.collect(Collectors.toList()));
		}
		return asUnmodifiable(getPrototypes(false));
	}

	/** {@inheritDoc} */
	@Override
	public UnmodifiableList<Prototype> getPrimaryPrototypes() {
		if(tutorialService.isTutorialActive()) {
			return asUnmodifiable(getPrototypes(true).stream()
					.filter(proto -> tutorialService.prototypeActive(proto))
					.collect(Collectors.toList()));
		}
		return asUnmodifiable(getPrototypes(true));
	}

	/** Returns prototypes including or excluding non-primaries. */
	private List<Prototype> getPrototypes(boolean primaryOnly) {
		List<Prototype> result = new ArrayList<Prototype>();
		for(IPrototypeProvider provider : prototypeProviderList) {
			for(Prototype p : provider.getPrototypes()) {
				if(!primaryOnly || p.isPrimary()) {
					result.add(p);
				}
			}
		}
		return result;
	}

	/** {@inheritDoc} */
	@Override
	public String getUniquePrototypeID(Prototype prototype) {
		for(IPrototypeProvider prov : prototypeProviderList) {
			String uid = prov.getUniqueID(prototype);
			if(uid != null) {
				return uid;
			}
		}
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public Prototype getPrototypeByUniqueID(String id) {
		for(IPrototypeProvider prov : prototypeProviderList) {
			Prototype proto = prov.getPrototypeByID(id);
			if(proto != null) {
				return proto;
			}
		}
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public List<Prototype> getComposablePrototypes(Class<? extends EObject> modelElementType) {
		List<Prototype> composablePrototypes = composablePrototypesCache.get(modelElementType);
		if(composablePrototypes != null) {
			return composablePrototypes;
		}
		IElementCompositorService ecs = IElementCompositorService.getInstance();
		composablePrototypes = new LinkedList<Prototype>();
		for(Prototype prototype : getAllPrototypes()) {
			if(ecs.canComposePrototype(modelElementType, prototype)) {
				composablePrototypes.add(prototype);
			}
		}
		composablePrototypes = unmodifiableList(composablePrototypes);
		composablePrototypesCache.put(modelElementType, composablePrototypes);
		return composablePrototypes;
	}

	/** {@inheritDoc} */
	@Override
	public List<PrototypeCategory> getAllTopLevelPrototypesCategories() {
		List<PrototypeCategory> result = new ArrayList<PrototypeCategory>();

		for(IPrototypeProvider provider : prototypeProviderList) {
			for(PrototypeCategory currentCategory : provider.getCategories()) {
				if(currentCategory.getParentCategory() == null) {
					result.add(currentCategory);
				}
			}
		}

		Comparator<PrototypeCategory> lexicographicalComparator = (PrototypeCategory category1,
				PrototypeCategory category2) -> category1.getName().compareTo(category2.getName());
		sort(result, lexicographicalComparator);

		return result;
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionLabel() {
		return "Prototype Service";
	}

	/** {@inheritDoc} */
	@Override
	public List<IIntrospectionItem> getIntrospectionItems() {
		return emptyList();
	}

	/** {@inheritDoc} */
	@Override
	public boolean showInIntrospectionNavigation() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public IIntrospectionDetailsItem getDetailsItem() {
		return new PrototypeKISSDetailsItem(prototypeProviderList);
	}
}
